/* bf.c
******************************************************************************

   Yet another Brainfuck interpreter

   Author: Stephan Beyer
   
   Copyright (C) GPL, 2003, 2004 Stephan Beyer - s-beyer@gmx.net
  
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <stdio.h>
#include <stdlib.h> /* malloc, atoi */
#include <string.h> /* strlen */

#include <slang.h> /* S-Lang library */
#if 0
#include <curses.h> /* GNU NCurses */
#endif

#include "bf.h"
#include "errors.h"

/* global variables: */
char *array; /* the Brainfuck array */
char *p; /* array pointer */
Opt opt; /* bf options */
char *program;
/* We put the brainfuck code 'compressed' into memory before interpreting
 * it. for that we need our global variable program.
 * One program char is (containing of 8 bits) can be splitted into:
 *   - amount of Brainfuck operator - 1, 5 bits (max number: 32)
 *   - Brainfuck operator, 3 bits (see operations enum)
 *   - opBEGINs and opENDs mustn't be compressed (amount bits must be 00000b)
 */
int psize; /* program size (because there's no termination character) */

/* put <j> <c>-characters at position <i> into <program>
 * 1 <= j <= 32 ; return value is the new position */
int PutIntoProgram(char c, char j, int i)
{
	*(program+i) = ((j-1) << 3);
	switch(c)
	{
		case '+':
			*(program+i) |= opPLUS;
			break;
		case '-':
			*(program+i) |= opMINUS;
			break;
		case '<':
			*(program+i) |= opLESS;
			break;
		case '>':
			*(program+i) |= opGREATER;
			break;
		case '[':
			for(; j > 0; j--)
			{
				*(program+i) = opBEGIN;
				i++;
			}
			i--;
			break;
		case ']':
			for(; j > 0; j--)
			{
				*(program+i) = opEND;
				i++;
			}
			i--;
			break;
		case '.':
			*(program+i) |= opOUTPUT;
			break;
		case ',':
			*(program+i) |= opINPUT;
			break;
	}
	i++;
	return i;
}

/* read input file and write it into <program> */
void ReadProgram()
{
	int stage; /* read stage (to avoid double code) */
	FILE *fp;
	int i;

	/* file opening banana */
	fp = fopen(opt.filename, "r");
	DoAndErr(!fp, fopen, errFILE); 

	for(stage = 0; stage <= 1; stage++)
	{
		int c; /* fgetc()ed char */
		char oldc = 0;
		char j = 0;
		i = 0;

		rewind(fp);

		/* stage 0: determine program size -> alloc mem
		 * stage 1: write program into mem */
		while ((c = fgetc(fp)) != EOF)
		{
			if ((c == '+') || (c == '-')
			 || (c == '<') || (c == '>')
			 || (c == '.') || (c == ',')
			 || (c == '[') || (c == ']'))
			{
				if (stage && !oldc) /* stage 1: init oldc */
					oldc = c;
				if ((oldc != c) || (j >= 32))
				{
					if (!stage) /* stage 0 */
					{
						/* [] must not be compressed */
						if ((oldc == '[')
						 || (oldc == ']'))
							i += j - 1;
						i++; /* increment size */
					} else /* stage 1 */
						i = PutIntoProgram(oldc, j, i);
					j = 1; /* this char is there one time */
				}
				else
					j++; /* increment amount of one char */
				oldc = c;
			} 
		}

		if (!stage) /* stage 0 */
		{
			/* alloc memory for program */
			psize = i; /* save global psize */
			program = malloc(psize);
			DoAndErr(!program, malloc, errMEMORY);
		} else /* stage 1 */
			PutIntoProgram(oldc, j, i);
	}
	fclose(fp);
}

/* implementations of the ',' operation */
char GetInput()
{
	char c;

	if ((opt.inputmode >= 1)
	 && (opt.inputmode <= 4))
	{
		/* We init and reset the terminal on each input (and not
		 * once at start and end of program) to prevent bugs with
		 * bf output... */
		/* init slang terminal */
		DoAndErr(SLang_init_tty(-1, 0, 1) == -1, SLang_init_tty,
								errSLANG);
		/* get any key */
		if (opt.inputmode <= 2) /* 1 or 2 */
			c = SLang_getkey();
		else /* 3 or 4 */
		{ /* get key but no escape characters */
			int tmp;
			do {
				tmp = SLkp_getkey();
			} while(tmp > 255);
			c = (unsigned char)tmp;
		}
		
		if (c == '\r') /* slang returns \r, make \n */
			c = '\n';
		
		if (!(opt.inputmode % 2)) /* 2 and 4 */
		{
			putchar(c); /* print them */
			fflush(stdout);
		}
		
		/* and reset */
		SLang_reset_tty();
	}
	else /* opt.inputmode == 0 (default) */
		c = getchar();

	if (opt.null && (c == '\n')) /* translate \n to \0? */
		c = 0;
	return c;
}

void ShowCode(char c)
{
	if (opt.showinput)
		fprintf(stderr, "%c", c);
}

/* begins interpretion in program at position i end returns where it
 * ended (either at ] or eof - on eof it returns -1)  */
int Interprete (int i)
{
	int do_nothing = 0; /* > 0 => skip loops */
	
	while (i < psize)
	{
		int k, num;
		char c = *(program + i);
		num = (c >> 3) + 1;
		c &= 0x07;
		i++;

		if ((do_nothing) && (c != opEND)) {
			if (c == opBEGIN)
				do_nothing++;
			continue;
		}

		switch(c)
		{
			case opGREATER:
				if (p+num > array+opt.cells)
					ErrorMsg(errOUTOFRANGE, "Out of range! You wanted to '>' beyond the last cell. See -c option.");
				p += num;
				ShowCode('>');
				break;
			case opLESS:
				if (p-num < array)
					ErrorMsg(errOUTOFRANGE, "Out of range! You wanted to '<' below the first cell.\nTo solve, add some '>'s at the beginning, for example.");
				p -= num;
				ShowCode('<');
				break;
			case opPLUS:
				if ((!opt.wraparound) && (*p+num > 255))
					ErrorMsg(errWRAPAROUND, "Out of range! You wanted to '+' a 0xFF byte. See -w option.");
				*p += num;
				ShowCode('+');
				break;
			case opMINUS:
				if ((!opt.wraparound) && (*p-num < 0))
					ErrorMsg(errWRAPAROUND, "Out of range! You wanted to '-' a 0x00 byte. See -w option.");
				*p -= num;
				ShowCode('-');
				break;
			case opOUTPUT:
				for (k = 0; k < num; k++)
				{
					ShowCode('.');
					putchar(*p);
				}
				fflush(stdout);
				break;
			case opINPUT:
				for (k = 0; k < num; k++)
				{
					ShowCode(',');
					*p = GetInput();
				}
				break;
			case opBEGIN:
				ShowCode('[');

				if (!*p)
				{
					do_nothing++;
					Debug(skip loop)
					break;
				}
				Debug(begin loop);

				while(*p)
				{
					Debug(loop)
					k = Interprete(i);
				}
				i = k;
				Debug(loop ended);
				break;
			case opEND:
				if (do_nothing)
					do_nothing--;
				else {
					ShowCode(']');
					return i;
				}
				break;
		}
	}
	return -1;
}

void ResetArray()
{
	int i;
	array = malloc(opt.cells+1);
	DoAndErr(!array, malloc, errMEMORY);
	p = array;

	for (i = 0; i < opt.cells; i++)
		*(p++) = 0;

	p = array;
}

/* display usage information */
void Usage(char *bin)
{
	printf	("Yet another Brainfuck interpreter       version %s\n", VERSION);
	puts	("(C) 2003, 2004, Stephan Beyer, GPL, s-beyer@gmx.net\n");
	puts	("Usage information: ");
	printf	("\t%s [-h] [options] inputfile\n\n", bin);
	puts	("Available options:");
	printf	("\t-c<num>   specify number of cells [%d]\n", opt.cells);
	puts	("\t-d        show some debugging information (stderr)");
	puts	("\t-i        show used code input (stderr)");
	puts	("\t-n        translate input: 10 (\\n) to 0");
	puts	("\t-w        disallow decrementing 0 and incrementing 255");
	printf	("\t-,<mode>  set input mode: 0-4 [%d]\n", opt.inputmode);
	puts	("\nSee the bf(1) manpage for more information.");
	puts	("Have fun!");
}

void HandleOptions(int c, char **v)
/* too lazy to use getopt this time ;) */
{
	int i;
	/* setting defaults */
	opt.showinput = 0;
	opt.cells = CELLS;
	opt.inputmode = 4;
	opt.debug = 0;
	opt.null = 0;
	opt.wraparound = 1; /* allow by default */
	
	if (c <= 1) /* no filename given */
	{
		Usage(*v);
		ErrorMsg (errOPT, "No input file given.");
	}
	if (!strcmp(*(v+1), "-h")) /* is first argument '-h'? */
	{
		Usage(*v);
		exit(0);
	}
	
	for(i = 1; i < c-1; i++) /* options? */
	{
		if ((**(v+i) == '-')
		 && (strlen(*(v+i)) >= 2))
		{
			switch(*(*(v+i)+1))
			{
				case 'c':
					opt.cells = atoi(*(v+i)+2);
					break;
				case 'd':
					opt.debug = 1;
					break;
				case 'i':
					opt.showinput = 1;
					break;
				case 'n':
					opt.null = 1;
					break;
				case 'w':
					opt.wraparound = 0;
					break;
				case ',':
					opt.inputmode = atoi(*(v+i)+2);
					break;
				default:
					Usage(*v);
					printf("Unable to handle option %s - ",
						*(v+i));
					ErrorMsg(errOPT, "Unknown option.");
			}
		} else {
			Usage(*v);
			printf("Invalid argument %s - ", *(v+i));
			ErrorMsg(errOPT, "Invalid argument(s).");
		}
	}
	opt.filename = *(v+i); /* last option MUST be filename */
}

int main(int argc, char **argv)
{
	/* options */
	HandleOptions(argc, argv);
	/* init slang kp if mode 3 or 4 */
	if ((opt.inputmode == 3)
	 || (opt.inputmode == 4))
	{
		SLtt_get_terminfo();
		DoAndErr(SLkp_init() == -1,
			SLkp_init, errSLANG);
	}

	ReadProgram();
	ResetArray();

	/* reading and interpreting */
	if(Interprete(0) > -1)
		ErrorMsg(errBRACKET, "Unmatching brackets!");

	/* cleanup */
	free(array);
	free(program);
	return 0;
}

/* TODO list
 * - add missing features needed for Brainfuck contests, if there are some
 * - make it DOS-compatible?
 */
